/*******************************************************************************
+
+  LEDA 7.0  
+
+
+  _time.c
+
+
+  Copyright (c) 1995-2023
+  by Algorithmic Solutions Software GmbH
+  All rights reserved.
+ 
*******************************************************************************/


#include <LEDA/system/basic.h>

#include <time.h>

#if defined(__unix__)
#include <unistd.h>
#include <sys/time.h>
#include <sys/times.h>
#endif

#if defined(__win32__) || defined(__win64__)
#if defined(MAXINT)
#undef MAXINT
#endif
#include <windows.h>
#endif




LEDA_BEGIN_NAMESPACE

#if defined(__unix__)

double cpu_time() 
{ struct timespec tspec;
  clock_gettime(CLOCK_REALTIME,&tspec);
  return double(tspec.tv_sec) + 0.000000001 * double(tspec.tv_nsec);
}
  
float used_time()
{ 
#if defined(CLK_TCK)
  long clk_tck = CLK_TCK;
#elif defined(HZ)
  long clk_tck = HZ;
#else
  long clk_tck = sysconf(_SC_CLK_TCK);
#endif

  tms x;
  times(&x);
  return  float(x.tms_utime)/clk_tck;
}

float elapsed_time()
{ 
  // elapsed time since midnight

  timeval tv;
  gettimeofday(&tv,NULL);

  time_t now = tv.tv_sec;
  
  tm* tptr = localtime(&now);
  tptr->tm_hour = 0;
  tptr->tm_min = 0;
  tptr->tm_sec = 0;

  time_t t0 = mktime(tptr);

  double real_t  = (tv.tv_sec - t0) + double(tv.tv_usec)/1000000;

  return float(real_t);
}


#elif defined(__win32__) || defined(__win64__)

double get_process_time()
{ HANDLE hProcess = GetCurrentProcess();

  FILETIME ftCreate, ftExit, ftKernel, ftUser;
  GetProcessTimes(hProcess,&ftCreate,&ftExit,&ftKernel,&ftUser);

  double msec = (ldexp(ftKernel.dwHighDateTime,32) + 
                 ftKernel.dwLowDateTime)/10000;

  msec += ((ldexp(ftUser.dwHighDateTime,32) + ftUser.dwLowDateTime)/10000);

  return msec/1000;
}


double cpu_time()
{ 
  // elapsed time since sytem start

/* 
  // 16 msec resolution

  FILETIME ft;
  GetSystemTimeAsFileTime(&ft);
  double msec = (ldexp(ft.dwHighDateTime,32)+ft.dwLowDateTime)/10000.0;
  return msec/1000;
*/

  HANDLE hProcess = GetCurrentProcess();

  FILETIME ftCreate, ftExit, ftKernel, ftUser;
  GetProcessTimes(hProcess,&ftCreate,&ftExit,&ftKernel,&ftUser);

  double t1=(ldexp(ftUser.dwHighDateTime,32)+ftUser.dwLowDateTime)/10000.0;
  double t2=(ldexp(ftKernel.dwHighDateTime,32)+ftKernel.dwLowDateTime)/10000.0;

  return (t1 + t2)/1000;
}

float used_time() { return (float)cpu_time(); }

float elapsed_time() { 
  return  float(clock())/CLOCKS_PER_SEC;
}

#else

double cpu_time()    { return  double(clock())/CLOCKS_PER_SEC; }
float used_time()    { return  float(clock())/CLOCKS_PER_SEC; }
float elapsed_time() { return  float(clock())/CLOCKS_PER_SEC; }

#endif

double cpu_time(double& T)
{ double t = cpu_time() - T;
  T += t;
  return t;
}

float used_time(float& T)
{ float t = used_time() - T;
  T += t;
  return t;
}

float elapsed_time(float& T)
{ float t = elapsed_time() - T;
  T += t;
  return t;
}

LEDA_END_NAMESPACE
